#!/bin/python3

'''Expression nodes'''

#> Imports
from dataclasses import dataclass, field

from .bases import CausticASTNode
from .bases import BaseOperator, BaseUnaryOperator, BaseBinaryOperator
#</Imports

#> Header >/
__all__ = (# Unary
           'UPlus', 'UMinus', 'Increment', 'Decrement', 'BitInvert', 'LogNot',
           # Binary
           'AssignExpr',
           ## Arithmetic
           'Add', 'Sub', 'Mult', 'Div', 'Mod', 'MMul',
           ## Comparison
           'Equality', 'Inequality', 'LessThan', 'GreaterThan', 'LessThanOrEquality', 'GreaterThanOrEquality', 'Nullish',
           ### Aliases
           'EQ', 'NE', 'LT', 'GT', 'LE', 'GE',
           ## Logical
           'LogAnd', 'LogOr', 'LogXor',
           ## Bitwise
           'BitAnd', 'BitOr', 'BitXor', 'LShift', 'RShift',
           # Ternary
           'Ternary')

# Unary
class UPlus(BaseUnaryOperator):
    '''Represents a unary plus operator'''
    __slots__ = ()
class UMinus(BaseUnaryOperator):
    '''Represents a unary minus operator'''
    __slots__ = ()
class Increment(BaseUnaryOperator):
    '''Represents an increment operator'''
    __slots__ = ()
class Decrement(BaseUnaryOperator):
    '''Represents a decrement operator'''
    __slots__ = ()
class BitInvert(BaseUnaryOperator):
    '''Represents a bit inversion operator'''
    __slots__ = ()
class LogNot(BaseUnaryOperator):
    '''Represents a logical inversion operator'''
    __slots__ = ()

# Binary
class AssignExpr(BaseBinaryOperator):
    '''Represents an assignment-expression'''
    __slots__ = ()
## Arithmetic
class Add(BaseBinaryOperator):
    '''Represents an addition operator'''
    __slots__ = ()
class Sub(BaseBinaryOperator):
    '''Represents a subtraction operator'''
    __slots__ = ()
class Mult(BaseBinaryOperator):
    '''Represents a multiplication operator'''
    __slots__ = ()
class Div(BaseBinaryOperator):
    '''Represents a division operator'''
    __slots__ = ()
class Mod(BaseBinaryOperator):
    '''Represents a modulus operator'''
    __slots__ = ()
class MMul(BaseBinaryOperator):
    '''Represents a matrix-multiplication operator'''
    __slots__ = ()
## Comparison
class Equality(BaseBinaryOperator):
    '''Represents an equality comparison operator'''
    __slots__ = ()
EQ = Equality
class Inequality(BaseBinaryOperator):
    '''Represents an inequality comparison operator'''
    __Slots__ = ()
NE = Inequality
class LessThan(BaseBinaryOperator):
    '''Represents a less-than comparison operator'''
    __Slots__ = ()
LT = LessThan
class GreaterThan(BaseBinaryOperator):
    '''Represents a less-than comparison operator'''
    __Slots__ = ()
GT = GreaterThan
class LessThanOrEquality(BaseBinaryOperator):
    '''Represents a less-than or equality comparison operator'''
    __slots__ = ()
LE = LessThanOrEquality
class GreaterThanOrEquality(BaseBinaryOperator):
    '''Represents a greater-than or equality comparison operator'''
    __slots__ = ()
GE = GreaterThanOrEquality
class Nullish(BaseBinaryOperator):
    '''Represents a nullish comparison (nullish coalescing) operator'''
    __slots__ = ()
## Logical
class LogAnd(BaseBinaryOperator):
    '''Represents a logical and operator'''
    __slots__ = ()
class LogOr(BaseBinaryOperator):
    '''Represents a logical or operator'''
    __slots__ = ()
class LogXor(BaseBinaryOperator):
    '''Represents a logical exclusive-or operator'''
    __slots__ = ()
## Bitwise
class BitAnd(BaseBinaryOperator):
    '''Represents a bitwise and operator'''
    __slots__ = ()
class BitOr(BaseBinaryOperator):
    '''Represents a bitwise or operator'''
    __slots__ = ()
class BitXor(BaseBinaryOperator):
    '''Represents a bitwise exclusive or operator'''
    __slots__ = ()
class LShift(BaseBinaryOperator):
    '''Represents a left-shift operator'''
    __slots__ = ()
class RShift(BaseBinaryOperator):
    '''Represents a right-shift operator'''
    __slots__ = ()

# Ternary
@dataclass(slots=True, kw_only=True)
class Ternary(BaseOperator):
    '''Represents a ternary operator'''
    cmp: CausticASTNode = field(kw_only=False)
    true: CausticASTNode
    false: CausticASTNode



#!/bin/python3

'''The CST--Caustic (Abstract) Syntax Tree'''

#> Package >/
__all__ = ('CausticASTNode', 'bases',
           'atom', 'block', 'expression', 'statement',
           'mods')

from .bases import CausticASTNode
from . import bases

from . import atom
from . import block
from . import expression
from . import statement



#!/bin/python3

'''Block nodes, `Line` and `Block`'''

#> Imports
from dataclasses import dataclass
from collections import abc as cabc

from .bases import CausticASTNode
#</Imports

#> Header >/
__all__ = ('Line', 'Block')

_dc = dataclass(slots=True)

@_dc
class Line(CausticASTNode):
    '''Represents a single "line"--a statement, expression, etc.'''
    content: CausticASTNode
@_dc
class Block(CausticASTNode):
    '''Represents a block of lines'''
    body: cabc.Sequence[Line]



#!/bin/python3

'''
    CST node for atoms (literals, identifiers, etc.)
    Note that atoms are considered subtypes of expressions,
        A.E. an atom is an expression
'''

#> Imports
import codecs
from dataclasses import dataclass, field
from collections import abc as cabc

from .bases import BaseAtom, BaseLiteral
#</Imports

#> Header >/
__all__ = ('Identifier', 'DottedIdentifier',
           'Integer', 'Decimal',
           'Char', 'Bytes', 'String')

_dc = dataclass(slots=True)

@_dc
class Identifier(BaseAtom):
    '''Represents an identifier'''
    name: str

@_dc
class DottedIdentifier(BaseAtom):
    '''Represents a multipart identifier'''
    names: cabc.Sequence[str]

# Literals
## Numeric
@_dc
class Integer(BaseLiteral):
    '''Represents an integer literal'''
    val: str
    base: int = 10
@_dc
class Decimal(BaseLiteral):
    '''Represents a decimal'''
    val: str
## Character
@_dc
class Char(BaseLiteral):
    '''Represents a character'''
    val: str

@_dc
class Bytes(BaseLiteral):
    '''Represents a sequence of bytes'''
    val: bytes
    raw: bool = field(default=False, kw_only=True)
@_dc
class String(BaseLiteral):
    '''Represents a string'''
    val: bytes
    raw: bool = field(default=False, kw_only=True)



#!/bin/python3

'''CST nodes for an exception module'''

#> Imports
from collections import abc as cabc
from dataclasses import dataclass, field

from ..bases import CausticASTNode, BaseStatement
from ..block import Block
#</Imports

#> Header >/

__all__ = ('RaiseStatement', 'TryStatement', 'CatchStatement', 'LowerStatement')

_dc = dataclass(slots=True)

class RaiseStatement(BaseStatement):
    '''Represents an exception-raising statement'''
    target: CausticASTNode
class TryStatement(BaseStatement):
    '''Represents an exception-handling statement/block'''
    block: Block
    catch: 'CatchStatement' = field(kw_only=True)
class CatchStatement(BaseStatement):
    '''Represents the "catch" part of a `TryStatement`'''
    catches: cabc.Sequence[CausticASTNode]
    block: Block
class LowerStatement(BaseStatement):
    '''Represents an exception-ignoring statement/block'''
    lowers: cabc.Sequence[CausticASTNode]
    block: Block



#!/bin/python3

'''CST nodes for statements'''

#> Imports
import typing
from dataclasses import dataclass, field
from collections import abc as cabc

from .bases import CausticASTNode, BaseStatement
from .block import Block, Line
#</Imports

#> Header >/
__all__ = ('ImportStatement', 'ExportStatement',
           'IfStatement', 'ElifStatement', 'ElseStatement',
           'ForStatement', 'WhileStatement',
           'Directive')

_dc = dataclass(slots=True)

# External-related
@_dc
class ImportStatement(BaseStatement):
    '''Represents an import'''
    target: CausticASTNode
    i_from: CausticASTNode | None = field(default=None, kw_only=True)
    i_as:   CausticASTNode | None = field(default=None, kw_only=True)
@_dc
class ExportStatement(BaseStatement):
    '''Represents an export'''
    target: CausticASTNode
    e_as: str | None = field(default=None, kw_only=True)
    extern: bool = field(default=False, kw_only=True)

# Conditional-related
@_dc
class IfStatement(BaseStatement):
    '''Represents an if statement'''
    condition: CausticASTNode
    block: Block
    next: typing.Union['ElifStatement', 'ElseStatement'] | None = None
class ElifStatement(IfStatement):
    '''Represents an else-if statement'''
    __slots__ = ()
@_dc
class ElseStatement(BaseStatement):
    '''Represents an else statement'''
    block: Block

# Loop-related
@_dc
class ForStatement(BaseStatement):
    '''Represents a for loop'''
    clause: tuple[Line, Line, Line] | CausticASTNode
    block: Block
    for_in: bool = field(default=False, kw_only=True)
@_dc
class WhileStatement(BaseStatement):
    '''Represents a while loop'''
    clause: CausticASTNode
    block: Block
    do_while: bool = field(default=False, kw_only=True)

# Directive
class Directive(BaseStatement):
    '''Represents a compiler directive'''
    name: str
    args: cabc.Sequence[typing.Any]



#!/bin/python3

'''Base nodes, should not be invoked directly'''

#> Imports
import typing
from dataclasses import dataclass
#</Imports

#> Header >/
__all__ = ('CausticASTNode',
           'BaseStatement',
           'BaseExpression',
               'BaseOperator', 'BaseBinaryOperator', 'BaseUnaryOperator',
               'BaseAtom', 'BaseLiteral')

_dc = dataclass(slots=True)

@dataclass(slots=True, kw_only=True)
class CausticASTNode:
    '''
        The base class for all CST nodes
        `metadata` is (optionally) set by the parser and could
            be anything (depending on the parser), most commonly
            source information (like name and line/column info)
    '''
    metadata: typing.Any | None = None

class BaseStatement(CausticASTNode):
    '''The base class for all statement CST nodes'''
    __slots__ = ()

# Expressions
class BaseExpression(CausticASTNode):
    '''The base class for all expression CST nodes'''
    __slots__ = ()

class BaseOperator(BaseExpression):
    '''The base class for all operator CST nodes'''
    __slots__ = ()
@_dc
class BaseBinaryOperator(BaseOperator):
    '''The base class for all binary operator CST nodes'''
    left: CausticASTNode
    right: CausticASTNode
@_dc
class BaseUnaryOperator(BaseOperator):
    '''The base class for all unary operator CST nodes'''
    target: CausticASTNode

## Atoms
class BaseAtom(BaseExpression):
    '''The base class for all atom CST nodes'''
    __slots__ = ()
class BaseLiteral(BaseAtom):
    '''The base class for all literal CST nodes'''
    __slots__ = ()
